@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2022 Artifex Software, Inc.
+/* Copyright (C) 2001-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -191,8 +191,10 @@ static void
 FF_stream_close(FT_Stream str)
 {
     stream *ps = (stream *) str->descriptor.pointer;
+    gs_memory_t *mem = ps->memory;
 
     (void)sclose(ps);
+    gs_free_object(mem, ps, "FF_stream_close");
 }
 
 extern const uint file_default_buffer_size;
@@ -1275,10 +1277,10 @@ gs_fapi_ft_get_scaled_font(gs_fapi_server * a_server, gs_fapi_font * a_font,
             /* It must be type 42 (see code in FAPI_FF_get_glyph in zfapi.c). */
             else {
                 /* Get the length of the TrueType data. */
-                unsigned long ms;
 
                 if (a_font->retrieve_tt_font != NULL) {
-                    code = a_font->retrieve_tt_font(a_font, &own_font_data, &ms);
+                    unsigned int ms;
+                    code = a_font->retrieve_tt_font(a_font, (void **)&own_font_data, &ms);
                     if (code == 0) {
                         data_owned = false;
                         open_args.memory_base = own_font_data;
@@ -1289,13 +1291,14 @@ gs_fapi_ft_get_scaled_font(gs_fapi_server * a_server, gs_fapi_font * a_font,
                     code = gs_error_unregistered;
 
                 if (code < 0) {
-                    code = a_font->get_long(a_font, gs_fapi_font_feature_TT_size, 0, &ms);
+                    unsigned long lms;
+                    code = a_font->get_long(a_font, gs_fapi_font_feature_TT_size, 0, &lms);
                     if (code < 0)
                         return code;
-                    if (ms == 0)
+                    if (lms == 0)
                         return_error(gs_error_invalidfont);
 
-                    open_args.memory_size = (FT_Long)ms;
+                    open_args.memory_size = (FT_Long)lms;
 
                     /* Load the TrueType data into a single buffer. */
                     open_args.memory_base = own_font_data =
