@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* PLANar devices */
@@ -393,7 +393,7 @@ plan_open(gx_device * pdev)
 #ifdef DEBUG_PRINT
     emprintf(pdev->memory, "plan_open\n");
 #endif
-    code = gdev_prn_open_planar(pdev, 1);
+    code = gdev_prn_open_planar(pdev, pdev->color_info.num_components);
     if (code < 0)
         return code;
     pdev->color_info.separable_and_linear = GX_CINFO_SEP_LIN;
