@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2022 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -134,6 +134,8 @@ static const gs_param_item_t pdf_param_items[] = {
     pi("OmitXMP", gs_param_type_bool, OmitXMP),
     pi("ModifiesPageSize", gs_param_type_bool, ModifiesPageSize),
     pi("ModifiesPageOrder", gs_param_type_bool, ModifiesPageOrder),
+    pi("WriteObjStms", gs_param_type_bool, WriteObjStms),
+    pi("WriteXRefStm", gs_param_type_bool, WriteXRefStm),
 #undef pi
     gs_param_item_end
 };
@@ -287,12 +289,12 @@ gdev_pdf_get_param(gx_device *dev, char *Param, void *list)
                 ocrstr.size = 5;
                 ocrstr.persistent = false;
                 break;
-            UseOCRAsNeeded:
+            case UseOCRAsNeeded:
                 ocrstr.data = (const byte *)"AsNeeded";
                 ocrstr.size = 8;
                 ocrstr.persistent = false;
                 break;
-            UseOCRAlways:
+            case UseOCRAlways:
                 ocrstr.data = (const byte *)"Always";
                 ocrstr.size = 8;
                 ocrstr.persistent = false;
@@ -345,12 +347,12 @@ gdev_pdf_get_params(gx_device * dev, gs_param_list * plist)
                 ocrstr.size = 5;
                 ocrstr.persistent = false;
                 break;
-            UseOCRAsNeeded:
+            case UseOCRAsNeeded:
                 ocrstr.data = (const byte *)"AsNeeded";
                 ocrstr.size = 8;
                 ocrstr.persistent = false;
                 break;
-            UseOCRAlways:
+            case UseOCRAlways:
                 ocrstr.data = (const byte *)"Always";
                 ocrstr.size = 8;
                 ocrstr.persistent = false;
@@ -470,11 +472,17 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         gs_param_string langstr;
         switch (code = param_read_string(plist, (param_name = "OCRLanguage"), &langstr)) {
             case 0:
-                len = langstr.size;
-                if (len >= sizeof(pdev->ocr_language))
-                    len = sizeof(pdev->ocr_language)-1;
-                memcpy(pdev->ocr_language, langstr.data, len);
-                pdev->ocr_language[len] = 0;
+                if (pdev->memory->gs_lib_ctx->core->path_control_active
+                && (strlen(pdev->ocr_language) != langstr.size || memcmp(pdev->ocr_language, langstr.data, langstr.size) != 0)) {
+                    return_error(gs_error_invalidaccess);
+                }
+                else {
+                    len = langstr.size;
+                    if (len >= sizeof(pdev->ocr_language))
+                        len = sizeof(pdev->ocr_language)-1;
+                    memcpy(pdev->ocr_language, langstr.data, len);
+                    pdev->ocr_language[len] = 0;
+                }
                 break;
             case 1:
                 break;
@@ -721,10 +729,16 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
      * legal parameter values for psdf_put_params varies according to
      * the version.
      */
-    if (pdev->PDFX)
+    if (pdev->PDFX) {
         cl = (float)1.3; /* Instead pdev->CompatibilityLevel = 1.2; - see below. */
-    if (pdev->PDFA == 1 && cl != 1.4)
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    }
+    if (pdev->PDFA == 1 && cl != 1.4) {
         cl = (float)1.4;
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    }
     if (pdev->PDFA == 2 && cl < 1.7)
         cl = (float)1.7;
     pdev->version = (cl < 1.2 ? psdf_version_level2 : psdf_version_ll3);
@@ -780,6 +794,16 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
             }
         }
     }
+    /* Record if the user has specified a custom profile, so we don't replace it
+       with a default when we change color space - see win_pr2_set_bpp()
+       Here, we only want to know *if* we have a user specced profile, actually
+       setting it will be handled by gdev_vector_put_params()
+     */
+    if (!pdev->user_icc) {
+        gs_param_string icc_pro_dummy;
+
+        pdev->user_icc = param_read_string(plist, "OutputICCProfile", &icc_pro_dummy) == 0;
+    }
     switch (pdev->params.ColorConversionStrategy) {
         case ccs_ByObjectType:
         case ccs_LeaveColorUnchanged:
@@ -790,24 +814,30 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
             break;
         case ccs_CMYK:
             pdf_set_process_color_model(pdev, 2);
-            ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-            if (ecode < 0)
-                goto fail;
+            if (!pdev->user_icc) {
+                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                if (ecode < 0)
+                    goto fail;
+            }
             break;
         case ccs_Gray:
             pdf_set_process_color_model(pdev,0);
-            ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-            if (ecode < 0)
-                goto fail;
+            if (!pdev->user_icc) {
+                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                if (ecode < 0)
+                    goto fail;
+            }
             break;
         case ccs_sRGB:
         case ccs_RGB:
             /* Only bother to do this if we didn't handle it above */
             if (!pdev->params.ConvertCMYKImagesToRGB) {
                 pdf_set_process_color_model(pdev,1);
-                ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
-                if (ecode < 0)
-                    goto fail;
+                if (!pdev->user_icc) {
+                    ecode = gsicc_init_device_profile_struct((gx_device *)pdev, NULL, 0);
+                    if (ecode < 0)
+                        goto fail;
+                }
             }
             break;
         default:
@@ -883,6 +913,35 @@ gdev_pdf_put_params_impl(gx_device * dev, const gx_device_pdf * save_dev, gs_par
         pdev->Linearise = false;
     }
 
+    /* Object streams and XRef streams */
+    if (pdev->is_ps2write) {
+        pdev->WriteObjStms = false;
+        pdev->WriteXRefStm = false;
+    } else {
+        if (pdev->Linearise)
+            pdev->WriteObjStms = false;
+    }
+    if (pdev->WriteObjStms && pdev->CompatibilityLevel < 1.5) {
+        emprintf(pdev->memory, "Can't use Object streams before PDF 1.5, ignoring WriteObjStms directive\n");
+        pdev->WriteObjStms = false;
+    }
+    if (pdev->WriteXRefStm && pdev->CompatibilityLevel < 1.5) {
+        emprintf(pdev->memory, "Can't use an XRef stream before PDF 1.5, ignoring WriteXRefStm directive\n");
+        pdev->WriteXRefStm = false;
+    }
+    if (pdev->WriteObjStms && !pdev->WriteXRefStm) {
+        emprintf(pdev->memory, "Can't use Object streams without XRef stream, ignoring WriteObjStms directive\n");
+        pdev->WriteObjStms = false;
+    }
+    if (pdev->WriteObjStms) {
+        if (pdev->next_id == 1)
+            pdev->doubleXref = true;
+        if (!pdev->doubleXref) {
+            emprintf(pdev->memory, "Already started writing output, too late to select WriteObjStms, ignoring directive\n");
+            pdev->WriteObjStms = false;
+        }
+    }
+
     if (pdev->FlattenFonts)
         pdev->PreserveTrMode = false;
     return 0;
