@@ -1,4 +1,4 @@
-/* Copyright (C) 2001-2021 Artifex Software, Inc.
+/* Copyright (C) 2001-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 
@@ -63,6 +63,9 @@
 /* Define the maximum size of a destination array string. */
 #define MAX_DEST_STRING 80
 
+/* Define the maximum number of objects stored in an ObjStm */
+#define MAX_OBJSTM_OBJECTS 200
+
 /* ================ Types and structures ================ */
 
 typedef struct pdf_base_font_s pdf_base_font_t;
@@ -503,7 +506,7 @@ typedef struct pdf_viewer_state_s {
     bool text_knockout; /* state.text_knockout */
     bool fill_overprint;
     bool stroke_overprint;
-    bool stroke_adjust; /* state.stroke_adjust */
+    int stroke_adjust; /* state.stroke_adjust */
     bool fill_used_process_color;
     bool stroke_used_process_color;
     gx_hl_saved_color saved_fill_color;
@@ -641,7 +644,7 @@ struct gx_device_pdf_s {
     gs_param_string OwnerPassword;
     gs_param_string UserPassword;
     uint KeyLength;
-    uint Permissions;
+    int32_t Permissions;
     uint EncryptionR;
     gs_param_string NoEncrypt;
     bool EncryptMetadata;
@@ -661,6 +664,7 @@ struct gx_device_pdf_s {
     /* Additional graphics state */
     bool fill_overprint, stroke_overprint;
     int rendering_intent;
+    bool user_icc;
     bool remap_fill_color, remap_stroke_color;
     gs_id halftone_id;
     gs_id transfer_ids[4];
@@ -696,11 +700,19 @@ struct gx_device_pdf_s {
      */
     pdf_temp_file_t streams;
     /*
-     * pictures holds graphic objects being accumulated between BP and EP.
-     * The object is moved to streams when the EP is reached: since BP and
-     * EP nest, we delete the object from the pictures file at that time.
+     * ObjStm holds a tream of objects being stored in an ObjStm, up to MAX_OBJSTM_OBJECTS,
+     * We keep a record of the offset of each object within the stream in ObjStmOffsets, so
+     * that later on we can write the offset within the stream for each object into the
+     * beginning of the ObjStm. NumObjStmOffsets is used to keep track of how many we have,
+     * and ObjStm_id is the id of the stream which we'll eventually store in the 'asides'
+     * file, containing the ObjStm.
      */
-    pdf_temp_file_t pictures;
+    pdf_temp_file_t ObjStm;
+    long ObjStm_id;
+    gs_offset_t *ObjStmOffsets;
+    int NumObjStmObjects;
+    bool doubleXref;
+
     /* ................ */
     long next_id;
     /* The following 3 objects, and only these, are allocated */
@@ -946,7 +958,7 @@ struct gx_device_pdf_s {
                                      * anything in the image processing routines.
                                      */
     float UserUnit;
-    pdf_OCR_usage UseOCR;                     /* Never, AsNeeded or Always */
+    pdf_OCR_usage UseOCR;           /* Never, AsNeeded or Always */
     gs_text_enum_t* OCRSaved;       /* Saved state of the text enumerator before rendering glyph bitmaps for later OCR */
     pdf_OCR_stage OCRStage;         /* Used to control a (sort of) state machine when using OCR to get a Unicode value for a glyph */
     int *OCRUnicode;                /* Used to pass back the Unicode value from the OCR engine to the text processing */
@@ -954,6 +966,14 @@ struct gx_device_pdf_s {
     gs_glyph OCR_glyph;             /* Passes the current glyph code from text processing to the image processing code when rendering glyph bitmaps for OCR */
     ocr_glyph_t *ocr_glyphs;        /* Records bitmaps and other data from text processing when doing OCR */
     gs_gstate **initial_pattern_states;
+    bool OmitInfoDate;              /* If true, do not emit CreationDate and ModDate in the Info dictionary and XMP Metadata (must not be true for PDF/X support) */
+    bool OmitXMP;                   /* If true, do not emit an XMP /Metadata block and do not reference it from the Catalog (must not be true for PDF/A output) */
+    bool OmitID;                    /* If true, do not emit a /ID array in the trailer dicionary (must not be true for encrypted files or PDF 2.0) */
+    bool ModifiesPageSize;          /* If true, the new PDF interpreter will not preserve *Box values (the media size has been modified, they will be incorrect) */
+    bool ModifiesPageOrder;         /* If true, the new PDF interpreter will not preserve Outlines or Dests, because they will refer to the wrong page number */
+    bool WriteXRefStm;              /* If true, (the default) use an XRef stream rather than an xref table */
+    bool WriteObjStms;              /* If true, (the default) store candidate objects in ObjStms rather than plain text in the PDF file. */
+    int64_t PendingOC;
 };
 
 #define is_in_page(pdev)\
@@ -1112,6 +1132,12 @@ int pdf_record_usage_by_parent(gx_device_pdf *const pdev, long resource_id, long
 long pdf_open_separate(gx_device_pdf * pdev, long id, pdf_resource_type_t type);
 long pdf_begin_separate(gx_device_pdf * pdev, pdf_resource_type_t type);
 
+/* functions used for ObjStm writing */
+int FlushObjStm(gx_device_pdf *pdev);
+int NewObjStm(gx_device_pdf *pdev);
+long pdf_open_separate_noObjStm(gx_device_pdf * pdev, long id, pdf_resource_type_t type);
+int pdf_end_separate_noObjStm(gx_device_pdf * pdev, pdf_resource_type_t type);
+
 /* Reserve object id. */
 void pdf_reserve_object_id(gx_device_pdf * pdev, pdf_resource_t *ppres, long id);
 
@@ -1269,6 +1295,7 @@ typedef struct pdf_lcvd_s {
     gx_device_memory *mask;
     gx_device_pdf *pdev;
     dev_t_proc_copy_color((*std_copy_color), gx_device);
+    dev_t_proc_copy_mono((*std_copy_mono), gx_device);
     dev_t_proc_fill_rectangle((*std_fill_rectangle), gx_device);
     dev_t_proc_close_device((*std_close_device), gx_device);
     dev_t_proc_get_clipping_box((*std_get_clipping_box), gx_device);
@@ -1276,6 +1303,7 @@ typedef struct pdf_lcvd_s {
     bool mask_is_empty;
     bool path_is_empty;
     bool mask_is_clean;
+    bool filled_trap;
     bool write_matrix;
     bool has_background;
     gs_matrix m;
@@ -1290,7 +1318,7 @@ typedef struct pdf_lcvd_s {
 
 int pdf_setup_masked_image_converter(gx_device_pdf *pdev, gs_memory_t *mem, const gs_matrix *m, pdf_lcvd_t **pcvd,
                                  bool need_mask, int x, int y, int w, int h, bool write_on_close);
-int pdf_dump_converted_image(gx_device_pdf *pdev, pdf_lcvd_t *cvd);
+int pdf_dump_converted_image(gx_device_pdf *pdev, pdf_lcvd_t *cvd, int for_pattern);
 void pdf_remove_masked_image_converter(gx_device_pdf *pdev, pdf_lcvd_t *cvd, bool need_mask);
 
 /* ------ Miscellaneous output ------ */
@@ -1580,4 +1608,14 @@ int pdf_from_string_to_text(gx_device_pdf *pdev);
 void pdf_close_text_contents(gx_device_pdf *pdev);
 
 int gdev_pdf_get_param(gx_device *dev, char *Param, void *list);
+
+int pdf_open_temp_file(gx_device_pdf *pdev, pdf_temp_file_t *ptf);
+int pdf_open_temp_stream(gx_device_pdf *pdev, pdf_temp_file_t *ptf);
+int pdf_close_temp_file(gx_device_pdf *pdev, pdf_temp_file_t *ptf, int code);
+
+/* exported by gdevpdfe.c */
+
+int pdf_xmp_write_translated(gx_device_pdf* pdev, stream* s, const byte* data, int data_length,
+    void(*write)(stream* s, const byte* data, int data_length));
+
 #endif /* gdevpdfx_INCLUDED */
