@@ -1,4 +1,4 @@
-/* Copyright (C) 2018-2022 Artifex Software, Inc.
+/* Copyright (C) 2018-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* dictionary handling for the PDF interpreter */
@@ -198,9 +198,16 @@ pdfi_dict_get2(pdf_context *ctx, pdf_dict *d, const char *Key1,
 {
     int code;
 
-    code = pdfi_dict_get(ctx, d, Key1, o);
-    if (code == gs_error_undefined)
+    /* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. Assume abbreviated names are shorter :-) */
+    if (strlen(Key1) < strlen(Key2)) {
+        code = pdfi_dict_get(ctx, d, Key1, o);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get(ctx, d, Key2, o);
+    } else {
         code = pdfi_dict_get(ctx, d, Key2, o);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get(ctx, d, Key1, o);
+    }
     return code;
 }
 
@@ -489,9 +496,16 @@ pdfi_dict_get_type2(pdf_context *ctx, pdf_dict *d, const char *Key1, const char
 {
     int code;
 
-    code = pdfi_dict_get_type(ctx, d, Key1, type, o);
-    if (code == gs_error_undefined)
+    /* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. Assume abbreviated names are shorter :-) */
+    if (strlen(Key1) < strlen(Key2)) {
+        code = pdfi_dict_get_type(ctx, d, Key1, type, o);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_type(ctx, d, Key2, type, o);
+    } else {
         code = pdfi_dict_get_type(ctx, d, Key2, type, o);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_type(ctx, d, Key1, type, o);
+    }
     return code;
 }
 
@@ -518,9 +532,16 @@ pdfi_dict_get_int2(pdf_context *ctx, pdf_dict *d, const char *Key1,
 {
     int code;
 
-    code = pdfi_dict_get_int(ctx, d, Key1, i);
-    if (code == gs_error_undefined)
+    /* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. Assume abbreviated names are shorter :-) */
+    if (strlen(Key1) < strlen(Key2)) {
+        code = pdfi_dict_get_int(ctx, d, Key1, i);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_int(ctx, d, Key2, i);
+    } else {
         code = pdfi_dict_get_int(ctx, d, Key2, i);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_int(ctx, d, Key1, i);
+    }
     return code;
 }
 
@@ -559,9 +580,16 @@ pdfi_dict_get_bool2(pdf_context *ctx, pdf_dict *d, const char *Key1,
 {
     int code;
 
-    code = pdfi_dict_get_bool(ctx, d, Key1, val);
-    if (code == gs_error_undefined)
+    /* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. Assume abbreviated names are shorter :-) */
+    if (strlen(Key1) < strlen(Key2)) {
+        code = pdfi_dict_get_bool(ctx, d, Key1, val);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_bool(ctx, d, Key2, val);
+    } else {
         code = pdfi_dict_get_bool(ctx, d, Key2, val);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_bool(ctx, d, Key1, val);
+    }
     return code;
 }
 
@@ -592,9 +620,16 @@ int pdfi_dict_get_number2(pdf_context *ctx, pdf_dict *d, const char *Key1, const
 {
     int code;
 
-    code = pdfi_dict_get_number(ctx, d, Key1, f);
-    if (code == gs_error_undefined)
+    /* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. Assume abbreviated names are shorter :-) */
+    if (strlen(Key1) < strlen(Key2)) {
+        code = pdfi_dict_get_number(ctx, d, Key1, f);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_number(ctx, d, Key2, f);
+    } else {
         code = pdfi_dict_get_number(ctx, d, Key2, f);
+        if (code == gs_error_undefined)
+            code = pdfi_dict_get_number(ctx, d, Key1, f);
+    }
     return code;
 }
 
@@ -914,18 +949,25 @@ int pdfi_dict_put_unchecked(pdf_context *ctx, pdf_dict *d, const char *Key, pdf_
         return code;
     pdfi_countup(key);
 
-    /* Nope, its a new Key */
     if (d->size > d->entries) {
-        /* We have a hole, find and use it */
-        for (i=0;i< d->size;i++) {
-            if (d->list[i].key == NULL) {
-                d->list[i].key = key;
-                d->list[i].value = value;
-                pdfi_countup(value);
-                d->entries++;
-                return 0;
+        int search_start = d->entries < 1 ? 0 : d->entries - 1;
+        do {
+            /* We have a hole, find and use it */
+            for (i = search_start; i < d->size; i++) {
+                if (d->list[i].key == NULL) {
+                    d->list[i].key = key;
+                    d->list[i].value = value;
+                    pdfi_countup(value);
+                    d->entries++;
+                    return 0;
+                }
             }
-        }
+            if (search_start == 0) {
+                /* This shouldn't ever happen, but just in case.... */
+                break;
+            }
+            search_start = 0;
+        } while(1);
     }
 
     new_list = (pdf_dict_entry *)gs_alloc_bytes(ctx->memory, (d->size + 1) * sizeof(pdf_dict_entry), "pdfi_dict_put reallocate dictionary key/values");
