@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* Functions to deal with PDF structure, such as retrieving
@@ -1111,7 +1111,7 @@ static int pdfi_doc_mark_the_outline(pdf_context *ctx, pdf_dict *outline)
     uint64_t dictsize;
     uint64_t index;
     pdf_name *Key = NULL;
-    double num;
+    double num = 0;
 
     /* Basically we only do /Count, /Title, /A, /C, /F
      * The /First, /Last, /Next, /Parent get written magically by pdfwrite
@@ -1137,16 +1137,22 @@ static int pdfi_doc_mark_the_outline(pdf_context *ctx, pdf_dict *outline)
     if (code < 0)
         goto exit;
 
-    if (code > 0) {
+    /* We can't rely on Count being present to indicate that the Outline has descendants
+     * see bug #707228. Instead, look for the presence of a /First key. If there is no count
+     * key, or it is 0, then assume the entry is closed.
+     */
+    {
         pdf_dict *current = NULL, *next = NULL;
-        int count = 0;
+        int count = 0, code1;
 
-        code = pdfi_dict_knownget_type(ctx, outline, "First", PDF_DICT, (pdf_obj **)&current);
-        if (code > 0) {
+        code1 = pdfi_dict_knownget_type(ctx, outline, "First", PDF_DICT, (pdf_obj **)&current);
+        if (code1 > 0) {
+            if (code <= 0)
+                pdfi_set_warning(ctx, 0, NULL, W_PDF_OUTLINECHILD_NO_COUNT, "pdfi_doc_mark_the_outline", NULL);
             count++;
             do {
-                code = pdfi_dict_knownget_type(ctx, current, "Next", PDF_DICT, (pdf_obj **)&next);
-                if (code > 0) {
+                code1 = pdfi_dict_knownget_type(ctx, current, "Next", PDF_DICT, (pdf_obj **)&next);
+                if (code1 > 0) {
                     pdfi_countdown(current);
                     current = next;
                     next = NULL;
@@ -1156,7 +1162,7 @@ static int pdfi_doc_mark_the_outline(pdf_context *ctx, pdf_dict *outline)
             } while (1);
             pdfi_countdown(current);
         }
-        if (num < 0)
+        if (num <= 0)
             count *= -1;
         pdfi_dict_put_int(ctx, tempdict, "Count", count);
     }
@@ -1602,6 +1608,13 @@ static int pdfi_doc_EmbeddedFiles(pdf_context *ctx)
     pdf_array *Names_array = NULL;
     pdf_array *Kids = NULL;
 
+    code = pdfi_dict_knownget_type(ctx, ctx->Root, "Collection", PDF_DICT, (pdf_obj **)&Names);
+    if (code < 0) goto exit;
+    if (code > 0) {
+        code = 0;
+        goto exit;
+    }
+
     code = pdfi_dict_knownget_type(ctx, ctx->Root, "Names", PDF_DICT, (pdf_obj **)&Names);
     if (code <= 0) goto exit;
 
@@ -1676,6 +1689,111 @@ static int pdfi_doc_AcroForm(pdf_context *ctx)
 }
 
 
+static int pdfi_doc_view(pdf_context *ctx)
+{
+    int code = 0;
+    pdf_dict *tempdict = NULL;
+    pdf_name *Mode = NULL;
+    pdf_obj *ActionDest = NULL;
+
+    code = pdfi_dict_knownget_type(ctx, ctx->Root, "PageMode", PDF_NAME, (pdf_obj **)&Mode);
+    if (code < 0)
+        return code;
+
+    if (code != 0) {
+        code = pdfi_dict_alloc(ctx, 1, &tempdict);
+        if (code < 0)
+            return code;
+
+        pdfi_countup(tempdict);
+
+        code = pdfi_dict_put(ctx, tempdict, "PageMode", (pdf_obj *)Mode);
+        if (code < 0)
+            goto exit;
+
+        code = pdfi_pdfmark_from_dict(ctx, tempdict, NULL, "DOCVIEW");
+        if (code < 0)
+            goto exit;
+        pdfi_countdown(tempdict);
+        tempdict = NULL;
+    }
+
+    code = pdfi_dict_knownget_type(ctx, ctx->Root, "PageLayout", PDF_NAME, (pdf_obj **)&Mode);
+    if (code < 0)
+        return code;
+
+    if (code != 0) {
+        code = pdfi_dict_alloc(ctx, 1, &tempdict);
+        if (code < 0)
+            return code;
+
+        pdfi_countup(tempdict);
+
+        code = pdfi_dict_put(ctx, tempdict, "PageLayout", (pdf_obj *)Mode);
+        if (code < 0)
+            goto exit;
+
+        code = pdfi_pdfmark_from_dict(ctx, tempdict, NULL, "DOCVIEW");
+        if (code < 0)
+            goto exit;
+        pdfi_countdown(tempdict);
+        tempdict = NULL;
+    }
+
+    code = pdfi_dict_knownget(ctx, ctx->Root, "OpenAction", &ActionDest);
+    if (code < 0)
+        return code;
+
+    if (code != 0) {
+        if (pdfi_type_of(ActionDest) == PDF_DICT) {
+            /* Dictionary means this is an action */
+            code = pdfi_dict_alloc(ctx, 1, &tempdict);
+            if (code < 0)
+                return code;
+
+            pdfi_countup(tempdict);
+
+            code = pdfi_dict_put(ctx, tempdict, "A", (pdf_obj *)ActionDest);
+            if (code < 0)
+                goto exit;
+
+            code = pdfi_pdfmark_modA(ctx, tempdict);
+            if (code < 0) goto exit;
+
+            code = pdfi_pdfmark_from_dict(ctx, tempdict, NULL, "DOCVIEW");
+        } else {
+            if (pdfi_type_of(ActionDest) == PDF_ARRAY) {
+                /* Array means this is a destination */
+                code = pdfi_dict_alloc(ctx, 1, &tempdict);
+                if (code < 0)
+                    return code;
+
+                pdfi_countup(tempdict);
+
+                code = pdfi_dict_put(ctx, tempdict, "Dest", (pdf_obj *)ActionDest);
+                if (code < 0)
+                    goto exit;
+                code = pdfi_pdfmark_modDest(ctx, tempdict);
+                if (code < 0)
+                    goto exit;
+                code = pdfi_pdfmark_from_dict(ctx, tempdict, NULL, "DOCVIEW");
+                if (code < 0)
+                    goto exit;
+            } else {
+                code = gs_note_error(gs_error_typecheck);
+                goto exit;
+            }
+        }
+    }
+
+exit:
+    pdfi_countdown(ActionDest);
+    pdfi_countdown(Mode);
+    pdfi_countdown(tempdict);
+    return code;
+}
+
+
 /* See pdf_main.ps/process_trailer_attrs()
  * Some of this stuff is about pdfmarks, and some of it is just handling
  * random things in the trailer.
@@ -1687,6 +1805,8 @@ int pdfi_doc_trailer(pdf_context *ctx)
     /* Can't do this stuff with no Trailer */
     if (!ctx->Trailer) {
         pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_TRAILER, "pdfi_doc_trailer", NULL);
+        if (ctx->args.pdfstoponwarning)
+            code = gs_note_error(gs_error_undefined);
         goto exit;
     }
 
@@ -1699,21 +1819,37 @@ int pdfi_doc_trailer(pdf_context *ctx)
                 goto exit;
         }
 
+        /* Handle Docview pdfmark stuff */
+        if (ctx->args.preservedocview) {
+            code = pdfi_doc_view(ctx);
+            if (code < 0) {
+                pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_VIEW, "pdfi_doc_view", NULL);
+                if (ctx->args.pdfstoponwarning)
+                    goto exit;
+            }
+        }
+
         /* Handle Info */
         code = pdfi_doc_Info(ctx);
         if (code < 0) {
             pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_INFO, "pdfi_doc_trailer", NULL);
-            if (ctx->args.pdfstoponerror)
+            /* pdfwrite will set a Fatal error when processing the DOCINFO if it has been
+             * told to create a PDF/A. the PDFA compatibility is 2, and the file info
+             * cannot be handled. In that case, abort immediately.
+             */
+            if (ctx->args.pdfstoponwarning || code == gs_error_Fatal)
                 goto exit;
         }
 
         /* Handle EmbeddedFiles */
         /* TODO: add a configuration option to embed or omit */
-        code = pdfi_doc_EmbeddedFiles(ctx);
-        if (code < 0) {
-            pdfi_set_warning(ctx, 0, NULL, W_PDF_BAD_EMBEDDEDFILES, "pdfi_doc_trailer", NULL);
-            if (ctx->args.pdfstoponerror)
-                goto exit;
+        if (ctx->args.preserveembeddedfiles) {
+            code = pdfi_doc_EmbeddedFiles(ctx);
+            if (code < 0) {
+                pdfi_set_warning(ctx, 0, NULL, W_PDF_BAD_EMBEDDEDFILES, "pdfi_doc_trailer", NULL);
+                if (ctx->args.pdfstoponwarning)
+                    goto exit;
+            }
         }
     }
 
@@ -1724,7 +1860,7 @@ int pdfi_doc_trailer(pdf_context *ctx)
     code = pdfi_doc_AcroForm(ctx);
     if (code < 0) {
         pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_ACROFORM, "pdfi_doc_trailer", NULL);
-        if (ctx->args.pdfstoponerror)
+        if (ctx->args.pdfstoponwarning)
             goto exit;
     }
 
@@ -1732,7 +1868,7 @@ int pdfi_doc_trailer(pdf_context *ctx)
     code = pdfi_doc_OutputIntents(ctx);
     if (code < 0) {
         pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_OUTPUTINTENTS, "pdfi_doc_trailer", NULL);
-        if (ctx->args.pdfstoponerror)
+        if (ctx->args.pdfstoponwarning)
             goto exit;
     }
 
@@ -1740,7 +1876,7 @@ int pdfi_doc_trailer(pdf_context *ctx)
     code = pdfi_doc_PageLabels(ctx);
     if (code < 0) {
         pdfi_set_warning(ctx, code, NULL, W_PDF_BAD_PAGELABELS, "pdfi_doc_trailer", NULL);
-        if (ctx->args.pdfstoponerror)
+        if (ctx->args.pdfstoponwarning)
             goto exit;
     }
 
