@@ -1,4 +1,4 @@
-/* Copyright (C) 2020-2022 Artifex Software, Inc.
+/* Copyright (C) 2020-2024 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* common code for Postscript-type font handling */
@@ -117,30 +117,6 @@ pdf_ps_end_number_object(int c)
     return (c != '.' && c != 'e' && c != '-' && (c < '0' || c > '9'));
 }
 
-static inline bool
-ishex(char c)
-{
-    if (c < 0x30)
-        return false;
-
-    if (c > 0x39) {
-        if (c > 'F') {
-            if (c < 'a')
-                return false;
-            if (c > 'f')
-                return false;
-            return true;
-        }
-        else {
-            if (c < 'A')
-                return false;
-            return true;
-        }
-    }
-    else
-        return true;
-}
-
 static inline int
 decodehex(char c)
 {
@@ -225,15 +201,27 @@ pdfi_pscript_interpret(pdf_ps_ctx_t *cs, byte *pdfpsbuf, int64_t buflen)
                     while (pdfpsbuf < buflim && *pdfpsbuf != '>')
                         pdfpsbuf++;
                     len = pdfpsbuf - s;
-                    while (len % 2)
-                        len--;
                     for (i = 0; i < len; i += 2) {
-                        hbuf[0] = s[i];
-                        hbuf[1] = s[i + 1];
+                        /* If it's a single digit, see iff we can treat it as a nibble */
+                        if (i + 2 > len) {
+                            hbuf[0] = (byte)'0';
+                            hbuf[1] = s[i];
+                            len++;
+                        }
+                        else {
+                            hbuf[0] = s[i];
+                            hbuf[1] = s[i + 1];
+                        }
+                        if (!ishex(hbuf[0]) || !ishex(hbuf[1])) {
+                            i = len;
+                            break;
+                        }
                         *s2++ = (decodehex(hbuf[0]) << 4) | decodehex(hbuf[1]);
                     }
-                    pdfpsbuf++; /* move past the trailing '>' */
-                    code = pdf_ps_stack_push_string(cs, s, len >> 1);
+                    if (i >= len) {
+                        pdfpsbuf++; /* move past the trailing '>' */
+                        code = pdf_ps_stack_push_string(cs, s, len >> 1);
+                    }
                 }
                 break;
             case '>': /* For hex strings, this should be handled above */
@@ -466,7 +454,13 @@ ps_font_def_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
               }
               else if (pdf_ps_name_cmp(&s->cur[-1], "WMode")) {
                   if (pdf_ps_obj_has_type(&s->cur[0], PDF_PS_OBJ_INTEGER)) {
-                      priv->gsu.gst1.WMode = s->cur[0].val.i;
+                      if (s->cur[0].val.i != 0) {
+                          if (s->cur[0].val.i != 1)
+                            pdfi_set_warning(s->pdfi_ctx, 0, NULL, W_PDF_BAD_WMODE, "ps_font_def_func", NULL);
+                        priv->gsu.gst1.WMode = 1;
+                      }
+                      else
+                        priv->gsu.gst1.WMode = 0;
                   }
               }
               else if (pdf_ps_name_cmp(&s->cur[-1], "lenIV")) {
@@ -583,7 +577,7 @@ ps_font_def_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
                       if (code >= 0) {
                           pdfi_countup(pname);
 
-                          code = pdfi_create_Encoding(s->pdfi_ctx, (pdf_obj *) pname, NULL, (pdf_obj **) &new_enc);
+                          code = pdfi_create_Encoding(s->pdfi_ctx, NULL, (pdf_obj *) pname, NULL, (pdf_obj **) &new_enc);
                           if (code >= 0) {
                               pdfi_countdown(priv->u.t1.Encoding);
                               priv->u.t1.Encoding = new_enc;
@@ -885,7 +879,7 @@ ps_font_def_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
                                       }
                                   }
                                   if (code >= 0) {
-                                      pdfi_array_put(s->pdfi_ctx, priv->u.t1.blenddesignpositions, i, (pdf_obj *)sa);
+                                      code = pdfi_array_put(s->pdfi_ctx, priv->u.t1.blenddesignpositions, i, (pdf_obj *)sa);
                                   }
                                   pdfi_countdown(sa);
                               }
@@ -1045,21 +1039,14 @@ ps_font_array_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
         !memcmp(s->cur[-1].val.name, PDF_PS_OPER_NAME_AND_LEN("Subrs"))) {
 
         if (s->cur[0].val.i > 0) {
-            if (priv->u.t1.Subrs != NULL) {
-                int i;
-                for (i = 0; i < priv->u.t1.NumSubrs; i++) {
-                    gs_free_object(mem, priv->u.t1.Subrs[i].data, "ps_font_array_func(Subrs[i])");
-                }
-                gs_free_object(mem, priv->u.t1.Subrs, "ps_font_array_func(Subrs)");
-            }
+            pdfi_countdown(priv->u.t1.Subrs);
 
-            priv->u.t1.Subrs = (gs_string *) gs_alloc_bytes(mem, s->cur[0].val.i *sizeof(gs_string), "ps_font_array_func(Subrs)");
-            if (priv->u.t1.Subrs == NULL) {
-                return_error(gs_error_VMerror);
+            code = pdfi_object_alloc(s->pdfi_ctx, PDF_ARRAY, (unsigned int)s->cur[0].val.i, (pdf_obj **)&priv->u.t1.Subrs);
+            if (code < 0) {
+                return code;
             }
-            memset(priv->u.t1.Subrs, 0x00, s->cur[0].val.i * sizeof(gs_string));
+            pdfi_countup(priv->u.t1.Subrs);
         }
-        priv->u.t1.NumSubrs = s->cur[0].val.i;
         code = pdf_ps_stack_pop(s, 1);
     }
     else if (pdf_ps_obj_has_type(&s->cur[-1], PDF_PS_OBJ_NAME) &&
@@ -1162,14 +1149,18 @@ pdf_ps_RD_oper_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
                 size = s->cur[0].val.i;
                 buf++;
                 if (buf + size < bufend) {
-                    priv->u.t1.Subrs[inx].data =
-                        gs_alloc_bytes(mem, size, "pdf_ps_RD_oper_func(subr string)");
-                    if (priv->u.t1.Subrs[inx].data == NULL) {
-                        (void)pdf_ps_stack_pop(s, 2);
-                        return_error(gs_error_VMerror);
+                    pdf_string *subr_str;
+
+                    code = pdfi_object_alloc(s->pdfi_ctx, PDF_STRING, (unsigned int)size, (pdf_obj **)&subr_str);
+                    if (code < 0) {
+                        return code;
                     }
-                    memcpy(priv->u.t1.Subrs[inx].data, buf, size);
-                    priv->u.t1.Subrs[inx].size = size;
+                    memcpy(subr_str->data, buf, size);
+                    pdfi_countup(subr_str);
+                    code = pdfi_array_put(s->pdfi_ctx, priv->u.t1.Subrs, inx, (pdf_obj *)subr_str);
+                    pdfi_countdown(subr_str);
+                    if (code < 0)
+                        return code;
                 }
             }
         }
@@ -1239,6 +1230,12 @@ pdf_ps_put_oper_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
     return code;
 }
 
+static int
+pdf_ps_closefile_oper_func(gs_memory_t *mem, pdf_ps_ctx_t *s, byte *buf, byte *bufend)
+{
+    return bufend - buf;
+}
+
 static pdf_ps_oper_list_t ps_font_oper_list[] = {
     {PDF_PS_OPER_NAME_AND_LEN("RD"), pdf_ps_RD_oper_func},
     {PDF_PS_OPER_NAME_AND_LEN("-|"), pdf_ps_RD_oper_func},
@@ -1263,6 +1260,7 @@ static pdf_ps_oper_list_t ps_font_oper_list[] = {
     {PDF_PS_OPER_NAME_AND_LEN("for"), pdf_ps_pop4_oper_func},
     {PDF_PS_OPER_NAME_AND_LEN("put"), pdf_ps_put_oper_func},
     {PDF_PS_OPER_NAME_AND_LEN("StandardEncoding"), pdf_ps_standardencoding_oper_func},
+    {PDF_PS_OPER_NAME_AND_LEN("closefile"), pdf_ps_closefile_oper_func},
     {NULL, 0, NULL}
 };
 
@@ -1282,7 +1280,7 @@ pdfi_read_ps_font(pdf_context *ctx, pdf_dict *font_dict, byte *fbuf, int fbuflen
        and that can end up in a stackoverflow error, even though we have a complete font. Override it
        and let the Type 1 specific code decide for itself if it can use the font.
      */
-    if (code == gs_error_stackoverflow)
+    if (code == gs_error_pdf_stackoverflow)
         code = 0;
 
     return code;
