@@ -1,4 +1,4 @@
-/* Copyright (C) 2018-2022 Artifex Software, Inc.
+/* Copyright (C) 2018-2023 Artifex Software, Inc.
    All Rights Reserved.
 
    This software is provided AS-IS with no warranty, either express or
@@ -9,8 +9,8 @@
    of the license contained in the file LICENSE in this distribution.
 
    Refer to licensing information at http://www.artifex.com or contact
-   Artifex Software, Inc.,  1305 Grant Avenue - Suite 200, Novato,
-   CA 94945, U.S.A., +1(415)492-9861, for further information.
+   Artifex Software, Inc.,  39 Mesa Street, Suite 108A, San Francisco,
+   CA 94129, USA, for further information.
 */
 
 /* function creation for the PDF interpreter */
@@ -157,6 +157,15 @@ pdfi_parse_type4_func_stream(pdf_context *ctx, pdf_c_stream *function_stream, in
         if (c < 0)
             break;
         switch(c) {
+            case '%':
+                do {
+                    c = pdfi_read_byte(ctx, function_stream);
+                    if (c < 0)
+                        break;
+                    if (c == 0x0a || c == 0x0d)
+                        break;
+                }while (1);
+                break;
             case 0x20:
             case 0x0a:
             case 0x0d:
@@ -258,7 +267,7 @@ pdfi_parse_type4_func_stream(pdf_context *ctx, pdf_c_stream *function_stream, in
                         if (Op->length > Size)
                             return_error(gs_error_undefined);
 
-                        if (memcmp(Op->op, TokenBuffer, Size - 1) == 0)
+                        if (memcmp(Op->op, TokenBuffer, Size) == 0)
                             break;
                     }
                     if (i > NumOps)
@@ -296,7 +305,7 @@ pdfi_build_function_4(pdf_context *ctx, gs_function_params_t * mnDR,
     params.ops.data = 0;	/* in case of failure */
     params.ops.size = 0;	/* ditto */
 
-    if (function_obj->type != PDF_STREAM)
+    if (pdfi_type_of(function_obj) != PDF_STREAM)
         return_error(gs_error_undefined);
     Length = pdfi_stream_length(ctx, (pdf_stream *)function_obj);
 
@@ -378,7 +387,7 @@ pdfi_build_function_0(pdf_context *ctx, gs_function_params_t * mnDR,
     params.Size = params.array_step = params.stream_step = NULL;
     params.Order = 0;
 
-    if (function_obj->type != PDF_STREAM)
+    if (pdfi_type_of(function_obj) != PDF_STREAM)
         return_error(gs_error_undefined);
 
     code = pdfi_dict_from_obj(ctx, (pdf_obj *)function_obj, &function_dict);
@@ -444,7 +453,7 @@ pdfi_build_function_0(pdf_context *ctx, gs_function_params_t * mnDR,
 
     code = pdfi_make_int_array_from_dict(ctx, (int **)&params.Size, function_dict, "Size");
     if (code != params.m) {
-        if (code > 0)
+        if (code >= 0)
             code = gs_error_rangecheck;
         goto function_0_error;
     }
@@ -568,8 +577,10 @@ pdfi_build_function_3(pdf_context *ctx, gs_function_params_t * mnDR,
     if (code < 0)
         goto function_3_error;
 
-    if (code != 2 * params.k)
+    if (code != 2 * params.k) {
+        code = gs_note_error(gs_error_rangecheck);
         goto function_3_error;
+    }
     code = 0;
 
     for (i = 0; i < params.k; ++i) {
@@ -620,6 +631,7 @@ static int pdfi_build_sub_function(pdf_context *ctx, gs_function_t ** ppfn, cons
     int64_t Type;
     gs_function_params_t params;
     pdf_dict *stream_dict;
+    int obj_num;
 
     params.Range = params.Domain = NULL;
 
@@ -627,10 +639,11 @@ static int pdfi_build_sub_function(pdf_context *ctx, gs_function_t ** ppfn, cons
     if (code < 0)
         return code;
 
-    if (stream_obj->object_num != 0) {
-        if (pdfi_loop_detector_check_object(ctx, stream_obj->object_num))
+    obj_num = pdf_object_num(stream_obj);
+    if (obj_num != 0) {
+        if (pdfi_loop_detector_check_object(ctx, obj_num))
             return gs_note_error(gs_error_circular_reference);
-        code = pdfi_loop_detector_add_object(ctx, stream_obj->object_num);
+        code = pdfi_loop_detector_add_object(ctx, obj_num);
         if (code < 0)
             goto sub_function_error;
     }
